/*
* Copyright (c) 2018 Xilinx Inc
*
* This file is part of FFmpeg.
*
* FFmpeg is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* FFmpeg is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "libavutil/internal.h"
#include "libavutil/common.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"
#include "libavutil/imgutils.h"
#include "libavutil/timestamp.h"
#include "libavutil/fifo.h"
#include "avcodec.h"
#include "internal.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <xma.h>
#include <xrm.h>
#include <dlfcn.h>

#include "../xmaPropsTOjson.h"
#define PKT_FIFO_SIZE  25

#define MAX_DEC_PARAMS 11

typedef struct mpsoc_vcu_dec_ctx {
    const AVClass *class;
    XmaDecoderSession *dec_session;
    XmaParameter dec_params[MAX_DEC_PARAMS];
    xrmContext       *xrm_ctx;
    xrmCuListResource  decode_cu_list_res;
    int                decode_res_inuse;
    XmaDataBuffer  buffer;
    XmaFrame  xma_frame;
    XmaFrameProperties props;
    AVCodecContext *avctx;
    bool flush_sent;
    enum AVPixelFormat pix_fmt;
    uint32_t bitdepth;
    uint32_t codec_type;
    uint32_t low_latency;
    uint32_t entropy_buffers_count;
    uint32_t latency_logging;
    uint32_t splitbuff_mode;
    int first_idr_found;
    AVFifoBuffer *pkt_fifo;
} mpsoc_vcu_dec_ctx;

#define VD AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_DECODING_PARAM
#define OFFSET(x) offsetof(mpsoc_vcu_dec_ctx, x)

static int vcu_dec_get_out_buffer(struct AVCodecContext *s, AVFrame *frame, int flags);
#define FLAGS (AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM)

static const AVOption options[] = {
    { "pixel_format", "set output pixel format", OFFSET(pix_fmt), AV_OPT_TYPE_PIXEL_FMT, {.i64 = AV_PIX_FMT_VCU_NV12}, -1, INT_MAX, VD },
    { "low_latency", "Should low latency decoding be used", OFFSET(low_latency), AV_OPT_TYPE_INT, { }, 0, 1, VD, "low_latency" },
    { "entropy_buffers_count", "Specify number of internal entropy buffers", OFFSET(entropy_buffers_count), AV_OPT_TYPE_INT , { .i64 = 2 }, 2, 10, VD, "entropy_buffers_count" },
    { "latency_logging", "Log latency information to syslog", OFFSET(latency_logging), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 1, VD, "latency_logging" },
    { "splitbuff_mode", "configure decoder in split/unsplit input buffer mode", OFFSET(splitbuff_mode), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 1, VD, "splitbuff_mode" },
    { NULL },
};

static bool mpsoc_decode_is_h264_idr (AVPacket *pkt)
{
    unsigned char* pt = pkt->data;
    unsigned char* end = pkt->data + pkt->size - 3;
    while (pt < end)
    {
        if ((pt[0] == 0x00) && (pt[1] == 0x00) && (pt[2] == 0x01) && ((pt[3] & 0x1F) == 0x05))
            return true;
        pt++;
    }
    return false;
}

static bool mpsoc_decode_is_hevc_idr (AVPacket *pkt)
{
    unsigned char* pt = pkt->data;
    unsigned char* end = pkt->data + pkt->size - 3;
    while (pt < end)
    {
        if ((pt[0] == 0x00) && (pt[1] == 0x00) && (pt[2] == 0x01))
        {
            unsigned char naluType = (pt[3] & 0x7E) >> 1;
            if (naluType == 19 || naluType == 20 || naluType == 21)
                return true;
        }
        pt++;
    }
    return false;
}

static void  mpsoc_vcu_flush(AVCodecContext *avctx)
{
    mpsoc_vcu_dec_ctx *ctx = avctx->priv_data;

    /* reinitialize as we loop (-stream_loop) without going through init */
    ctx->flush_sent = false;
}

static av_cold int mpsoc_vcu_decode_close (AVCodecContext *avctx)
{
    mpsoc_vcu_dec_ctx *ctx = avctx->priv_data;

    if (ctx->pkt_fifo)
      av_fifo_free(ctx->pkt_fifo);


    if (getenv("XRM_RESERVE_ID"))
    {
       int xrm_reserve_id = atoi(getenv("XRM_RESERVE_ID"));

       //XRM encoder de-allocation     
       if (ctx->decode_res_inuse ==1)
       {
          if (!(xrmCuListRelease(ctx->xrm_ctx, &ctx->decode_cu_list_res)))       
             printf("fail to release decode HW cu\n");
       }
    }
    if (xrmDestroyContext(ctx->xrm_ctx) != XRM_SUCCESS)
       printf("XRM : destroy context failed\n");

    xma_dec_session_destroy(ctx->dec_session);
    return 0;
}

static int vcu_dec_get_out_buffer(struct AVCodecContext *s, AVFrame *frame, int flags)
{
    int size;
    uint16_t stride_wd;
    uint16_t stride_ht;
    mpsoc_vcu_dec_ctx *ctx;

    assert(c->codec_type == AVMEDIA_TYPE_VIDEO);
    if (!s || !frame)
        return -1;

    ctx  = s->priv_data;

    stride_wd = FFALIGN(s->width, 256);
    stride_ht = FFALIGN(s->height, 64);

    frame->width = s->width;;
    frame->height = s->height;
    frame->linesize[0] = stride_wd;
    frame->linesize[1] = stride_wd;
    frame->format      = ctx->pix_fmt;

    if (frame->format == AV_PIX_FMT_NV12)
    {
        if (frame->buf[0] && frame->buf[1]) {
    	frame->data[0] = frame->buf[0]->data;
    	frame->data[1] = frame->buf[1]->data;
    	return 0;
        } else {
            return -1;
    }
    }
    if (frame->format == AV_PIX_FMT_VCU_NV12)
        return av_frame_clone_xma_frame (frame, &(ctx->xma_frame));
    return -1;
}

static int32_t mpsoc_send_data (
    mpsoc_vcu_dec_ctx *ctx,
    unsigned char *buf,
    uint32_t       size,
    int32_t        pts,
    int            is_eof
)
{
    int data_used;
    int offset    = 0;
    while (offset < size)
    {
        ctx->buffer.data.buffer = buf;
        ctx->buffer.alloc_size = size;
        ctx->buffer.is_eof = is_eof;
        ctx->buffer.pts = pts;

        int32_t ret = xma_dec_session_send_data(ctx->dec_session, &ctx->buffer, &data_used);
        if (ret != XMA_SUCCESS)
            return ret;

        offset += data_used;
        pts = -1; /* only first packet will carry pts */
    }
    return XMA_SUCCESS;
}

static void _allocate_xrm_dec_cu(mpsoc_vcu_dec_ctx *ctx, XmaDecoderProperties *dec_props)
{

    if (getenv("XRM_RESERVE_ID"))
    {
       int xrm_reserve_id = atoi(getenv("XRM_RESERVE_ID"));
       int ret = -1;
       char pluginName[XRM_MAX_NAME_LEN]; 

       //create local XRM context
       ctx->xrm_ctx = (xrmContext *)xrmCreateContext(XRM_API_VERSION_1);
       if (ctx->xrm_ctx == NULL)
       {
          printf("create local XRM context failed\n");
          return;
       }
 
       //XRM decoder plugin load calculation
       int32_t func_id = 0, dec_load=0;

       xrmPluginFuncParam param;
       memset(&param, 0, sizeof(xrmPluginFuncParam));
       void *handle;
       void (*convertXmaPropsToJson)(void* props, char* funcName, char* jsonJob);

       handle = dlopen("/opt/xilinx/xrm/plugin/libxmaPropsTOjson.so", RTLD_NOW );
 
       convertXmaPropsToJson = dlsym(handle, "convertXmaPropsToJson");
       (*convertXmaPropsToJson) (dec_props, "DECODER", param.input);
       dlclose(handle);

       strcpy(pluginName, "xrmU30DecPlugin");
       if (xrmExecPluginFunc(ctx->xrm_ctx, pluginName, func_id, &param) != XRM_SUCCESS)
          printf("dec plugin function %d, fail to run the function\n", func_id);
       else
       {
          dec_load = atoi((char*)(strtok(param.output, " "))); 
          //printf("dec plugin function %d, success to run the function, input: %s, output: %s dec_load=%d\n", func_id, param.input, param.output, dec_load);
       }

       //XRM decoder allocation 
       xrmCuListProperty decode_cu_list_prop;  
    
       memset(&decode_cu_list_prop, 0, sizeof(xrmCuListProperty));
       memset(&ctx->decode_cu_list_res, 0, sizeof(xrmCuListResource));

       decode_cu_list_prop.cuNum = 2;
       strcpy(decode_cu_list_prop.cuProps[0].kernelName, "decoder");
       strcpy(decode_cu_list_prop.cuProps[0].kernelAlias, "DECODER_MPSOC");	
       decode_cu_list_prop.cuProps[0].devExcl = false;
       decode_cu_list_prop.cuProps[0].requestLoad = dec_load;
       decode_cu_list_prop.cuProps[0].poolId = xrm_reserve_id;

       strcpy(decode_cu_list_prop.cuProps[1].kernelName, "kernel_vcu_decoder");
       decode_cu_list_prop.cuProps[1].devExcl = false;
       decode_cu_list_prop.cuProps[1].requestLoad = 100;
       decode_cu_list_prop.cuProps[1].poolId = xrm_reserve_id;

       ret = xrmCuListAlloc(ctx->xrm_ctx, &decode_cu_list_prop, &ctx->decode_cu_list_res);
  
       if (ret != 0)
       {
          printf("xrm_allocation: fail to allocate cu list from reserve id %d\n", xrm_reserve_id );
          return;
       }
       else
       {
          ctx->decode_res_inuse =1;	
       }

       //Set XMA plugin SO and device index 
       dec_props->plugin_lib = ctx->decode_cu_list_res.cuResources[0].kernelPluginFileName;
       dec_props->dev_index = ctx->decode_cu_list_res.cuResources[0].deviceId;
       dec_props->ddr_bank_index = -1;//XMA to select the ddr bank based on xclbin meta data
       dec_props->cu_index = ctx->decode_cu_list_res.cuResources[1].cuId;
       dec_props->channel_id = ctx->decode_cu_list_res.cuResources[1].channelId;//SW kernel always used 100%     
       //printf("**********dec xrm pl=%s dev=%d u=%d ch=%d\n",dec_props->plugin_lib,dec_props->dev_index,dec_props->cu_index,dec_props->channel_id);
   }
   else
   {
       //Set XMA plugin SO and device index 
       dec_props->plugin_lib = "/opt/xilinx/xma_plugins/libvcu-xma-dec-plg.so";
       dec_props->dev_index = 0;
       dec_props->ddr_bank_index = -1;//XMA to select the ddr bank based on xclbin meta data
       dec_props->cu_index = 8;
       dec_props->channel_id = 0;//SW kernel always used 100%   
   } 
}

static av_cold int mpsoc_vcu_decode_init (AVCodecContext *avctx)
{
    XmaDecoderProperties dec_props;
    mpsoc_vcu_dec_ctx   *ctx  = avctx->priv_data;
    strcpy(dec_props.hwvendor_string, "MPSoC");

    dec_props.hwdecoder_type  = XMA_MULTI_DECODER_TYPE;
    dec_props.params          = ctx->dec_params;
    dec_props.param_cnt       = MAX_DEC_PARAMS;
    dec_props.width                 = avctx->width;
    dec_props.height                = avctx->height;
    dec_props.framerate.numerator   = avctx->framerate.num;

    /* some clips can make parser provide zero frame rate, correct the denominator to avoid
       division by zero in XMA and net framerate = 0.  TODO: What will be decoder load in XMA in this case ?
    */
    if (avctx->framerate.den)
        dec_props.framerate.denominator = avctx->framerate.den;
    else
        dec_props.framerate.denominator = 1;

   //printf("numerator=%d %f\n",  dec_props.framerate.numerator,  dec_props.framerate.numerator);

    uint32_t scan_type = avctx->field_order;

    /* bits_per_raw_sample can be zero from parser. In such case, rely on pix fmt
       TODO: 10bit support branch has proper bitdepth support. Need to be merged */

    if (avctx->bits_per_raw_sample) {
        ctx->bitdepth = avctx->bits_per_raw_sample;
    } else {
        if (avctx->pix_fmt == AV_PIX_FMT_YUV420P)
            ctx->bitdepth = 8;
        else {
            printf("unsupported pixel format\n");
            return AVERROR(EINVAL);
        }
    }

    avctx->pix_fmt = ctx->pix_fmt;
    ctx->avctx = avctx;
    ctx->flush_sent = false;

    uint32_t chroma_mode = 420;//default

    switch (ctx->pix_fmt) {
        case AV_PIX_FMT_NV12:
        case AV_PIX_FMT_VCU_NV12:
            chroma_mode = 420;
            break;
        default:
            printf("Invalid pixel format specified\n");
            return AVERROR(EINVAL);
    }
    ctx->dec_params[0].name   = "bitdepth";
    ctx->dec_params[0].type   = XMA_UINT32;
    ctx->dec_params[0].length = sizeof(ctx->bitdepth);
    ctx->dec_params[0].value  = &(ctx->bitdepth);

    ctx->codec_type = (avctx->codec_id == AV_CODEC_ID_H264) ? 0 : 1;
    ctx->dec_params[1].name   = "codec_type";
    ctx->dec_params[1].type   = XMA_UINT32;
    ctx->dec_params[1].length = sizeof(ctx->codec_type);
    ctx->dec_params[1].value  = &(ctx->codec_type);

    ctx->dec_params[2].name   = "low_latency";
    ctx->dec_params[2].type   = XMA_UINT32;
    ctx->dec_params[2].length = sizeof(ctx->low_latency);
    ctx->dec_params[2].value  = &(ctx->low_latency);

    ctx->dec_params[3].name   = "entropy_buffers_count";
    ctx->dec_params[3].type   = XMA_UINT32;
    ctx->dec_params[3].length = sizeof(ctx->entropy_buffers_count);
    ctx->dec_params[3].value  = &(ctx->entropy_buffers_count);

    uint32_t zero_copy = (avctx->pix_fmt == AV_PIX_FMT_VCU_NV12) ? 1 : 0;
    ctx->dec_params[4].name   = "zero_copy";
    ctx->dec_params[4].type   = XMA_UINT32;
    ctx->dec_params[4].length = sizeof(zero_copy);
    ctx->dec_params[4].value  = &(zero_copy);

    ctx->dec_params[5].name   = "profile";
    ctx->dec_params[5].type   = XMA_UINT32;
    ctx->dec_params[5].length = sizeof(avctx->profile);
    ctx->dec_params[5].value  = &(avctx->profile);

    ctx->dec_params[6].name   = "level";
    ctx->dec_params[6].type   = XMA_UINT32;
    ctx->dec_params[6].length = sizeof(avctx->level);
    ctx->dec_params[6].value  = &(avctx->level);

    ctx->dec_params[7].name   = "chroma_mode";
    ctx->dec_params[7].type   = XMA_UINT32;
    ctx->dec_params[7].length = sizeof(chroma_mode);
    ctx->dec_params[7].value  = &(chroma_mode);

    ctx->dec_params[8].name   = "scan_type";
    ctx->dec_params[8].type   = XMA_UINT32;
    ctx->dec_params[8].length = sizeof(scan_type);
    ctx->dec_params[8].value  = &(scan_type);

    ctx->dec_params[9].name   = "latency_logging";
    ctx->dec_params[9].type   = XMA_UINT32;
    ctx->dec_params[9].length = sizeof(ctx->latency_logging);
    ctx->dec_params[9].value  = &(ctx->latency_logging);

    ctx->dec_params[10].name   = "splitbuff_mode";
    ctx->dec_params[10].type   = XMA_UINT32;
    ctx->dec_params[10].length = sizeof(ctx->splitbuff_mode);
    ctx->dec_params[10].value  = &(ctx->splitbuff_mode);

    /*----------------------------------------------------
      Allocate decoder resource from XRM reserved resource
      ----------------------------------------------------*/
    ctx->decode_res_inuse    = 0;
    _allocate_xrm_dec_cu(ctx, &dec_props);

    ctx->dec_session = xma_dec_session_create(&dec_props);
    if (!ctx->dec_session) {
        printf("ERROR: Unable to allocate MPSoC decoder session.\n");
        return AVERROR_EXTERNAL;
    }

    switch (avctx->pix_fmt) {
        case AV_PIX_FMT_NV12:
            ctx->xma_frame.data[0].buffer = NULL;
            ctx->xma_frame.data[0].buffer_type = XMA_HOST_BUFFER_TYPE;
            ctx->xma_frame.data[1].buffer = NULL;
            ctx->xma_frame.data[1].buffer_type = XMA_HOST_BUFFER_TYPE;
            break;
        case AV_PIX_FMT_VCU_NV12:
            ctx->xma_frame.data[0].buffer = NULL;
            ctx->xma_frame.data[0].buffer_type = XMA_DEVICE_BUFFER_TYPE;
            ctx->xma_frame.data[1].buffer = NULL;
            ctx->xma_frame.data[1].buffer_type = XMA_HOST_BUFFER_TYPE;
            break;
    }
    for (uint32_t i = 0; i < 2; i++) {
        ctx->xma_frame.data[i].refcount = 1;
        ctx->xma_frame.data[i].is_clone = 1;
    }
    ctx->xma_frame.side_data = NULL;
    ctx->xma_frame.frame_props.format = XMA_VCU_NV12_FMT_TYPE;
    ctx->xma_frame.frame_props.width = avctx->width;
    ctx->xma_frame.frame_props.height = avctx->height;
    ctx->xma_frame.frame_props.bits_per_pixel = 8;
    ctx->xma_frame.frame_rate.numerator = avctx->framerate.num;
    ctx->xma_frame.frame_rate.denominator = avctx->framerate.den;


    ctx->pkt_fifo = av_fifo_alloc(PKT_FIFO_SIZE * sizeof(AVPacket));
    if (!ctx->pkt_fifo)
       return AVERROR(ENOMEM);
    return 0;
}

static int mpsoc_vcu_decode (AVCodecContext *avctx, void *data, int *got_frame, AVPacket *avpkt)
{
    int send_ret, recv_ret;

    AVFrame *frame	= data;
    int data_used	= 0;
    mpsoc_vcu_dec_ctx *ctx = avctx->priv_data;
    AVPacket *avpkt_clone = NULL;
    AVPacket buffer_pkt = { 0 };

    if (avpkt->data) {
        if (ctx->first_idr_found == 0)
        {
            if ((avctx->codec_id == AV_CODEC_ID_H264) ? mpsoc_decode_is_h264_idr (avpkt) : mpsoc_decode_is_hevc_idr (avpkt))
                ctx->first_idr_found = 1;
            else
            {
                *got_frame = 0;
                return avpkt->size;
            }
        }

        if (av_fifo_size(ctx->pkt_fifo)) {
             av_fifo_generic_peek(ctx->pkt_fifo, &buffer_pkt, sizeof(AVPacket ), NULL);
             avpkt_clone = av_packet_clone(avpkt);

             if (av_fifo_space(ctx->pkt_fifo) < sizeof(AVPacket )) {
                 int ret = av_fifo_grow(ctx->pkt_fifo, sizeof(AVPacket));
                 if (ret < 0)
                     return ret;
             }
             av_fifo_generic_write(ctx->pkt_fifo, avpkt_clone, sizeof(AVPacket ), NULL);

             send_ret = mpsoc_send_data(ctx, buffer_pkt.data, buffer_pkt.size, buffer_pkt.pts, 0);
        } else {
             send_ret = mpsoc_send_data(ctx, avpkt->data, avpkt->size, avpkt->pts, 0);
        }

        if (send_ret == XMA_ERROR) {
            *got_frame = 0;
            return AVERROR(EIO);
        } else {
            if (avctx->pix_fmt == AV_PIX_FMT_NV12)
            {
                uint32_t stride_wd = FFALIGN(avctx->width, 256);
                uint32_t stride_ht = FFALIGN(avctx->height, 64);

                frame->buf[0] = av_buffer_alloc (stride_wd*stride_ht);
                ctx->xma_frame.data[0].buffer = frame->buf[0]->data;
                frame->buf[1] = av_buffer_alloc (stride_wd*stride_ht/2);
                ctx->xma_frame.data[1].buffer = frame->buf[1]->data;
            }
            recv_ret = xma_dec_session_recv_frame(ctx->dec_session, &(ctx->xma_frame));
            if (recv_ret != XMA_SUCCESS) {
                if (avctx->pix_fmt == AV_PIX_FMT_NV12)
                {
                    av_buffer_unref(&frame->buf[0]);
                    av_buffer_unref(&frame->buf[1]);
                }
                *got_frame = 0;
            } else {
                if (vcu_dec_get_out_buffer(avctx, frame, 0)) {
                    *got_frame = 0;
                } else {
                    *got_frame = 1;
                    frame->pts = ctx->xma_frame.pts;
                }
            }

            if (send_ret == XMA_TRY_AGAIN) {
                if (!av_fifo_size(ctx->pkt_fifo)) {
                   avpkt_clone = av_packet_clone(avpkt);
                   av_fifo_generic_write(ctx->pkt_fifo, avpkt_clone, sizeof(AVPacket ), NULL);
             }

             } else {
                if (av_fifo_size(ctx->pkt_fifo)) {
                   av_fifo_generic_read(ctx->pkt_fifo, &buffer_pkt, sizeof(AVPacket ), NULL);
                   av_packet_unref(&buffer_pkt);
                }
             }

        }

        data_used =  avpkt->size;

    } else { // EOF

         while(true) {
                if (av_fifo_size(ctx->pkt_fifo)) {
                    av_fifo_generic_peek(ctx->pkt_fifo, &buffer_pkt, sizeof(AVPacket), NULL);
                    send_ret = mpsoc_send_data(ctx, buffer_pkt.data, buffer_pkt.size, buffer_pkt.pts, 0);
                } else {
                    /* flush not sent and queue is empty. send flush now */
                if (!ctx->flush_sent) {
            ctx->flush_sent = true;
            ctx->buffer.is_eof = 1;
                    send_ret = xma_dec_session_send_data(ctx->dec_session, &(ctx->buffer), &data_used);
                } else {
                /* send free output buffer indexes, so that decoding can continue on device side */
                  XmaDataBuffer eos_buff;
                  eos_buff.data.buffer = NULL;
                  eos_buff.alloc_size = 0;
                  eos_buff.is_eof = 0;
                  eos_buff.pts = -1;
                  send_ret = xma_dec_session_send_data(ctx->dec_session, &eos_buff, &data_used);
                }
            }

                if (send_ret == XMA_ERROR) {
                *got_frame = 0;
                return AVERROR_UNKNOWN;
            }
                if (send_ret == XMA_SUCCESS) {
                    if (av_fifo_size(ctx->pkt_fifo)) {
                        av_fifo_generic_read(ctx->pkt_fifo, &buffer_pkt, sizeof(AVPacket), NULL);
                        av_packet_unref(&buffer_pkt);
                    }
                }

            if (avctx->pix_fmt == AV_PIX_FMT_NV12)
            {
                uint32_t stride_wd = FFALIGN(avctx->width, 256);
                uint32_t stride_ht = FFALIGN(avctx->height, 64);

                frame->buf[0] = av_buffer_alloc (stride_wd*stride_ht);
                ctx->xma_frame.data[0].buffer = frame->buf[0]->data;
                frame->buf[1] = av_buffer_alloc (stride_wd*stride_ht/2);
                ctx->xma_frame.data[1].buffer = frame->buf[1]->data;
            }
            
	        recv_ret = xma_dec_session_recv_frame(ctx->dec_session, &(ctx->xma_frame));
                if (recv_ret != XMA_SUCCESS) {
                    if (avctx->pix_fmt == AV_PIX_FMT_NV12)
                    {
                        av_buffer_unref(&frame->buf[0]);
                        av_buffer_unref(&frame->buf[1]);
                }
                    *got_frame = 0;
                } else {
                    if (vcu_dec_get_out_buffer(avctx, frame, 0)) {
                        *got_frame = 0;
                    } else {
                    *got_frame = 1;
                        frame->pts = ctx->xma_frame.pts;
                    }
                }

           if (!ctx->flush_sent) {
             if (recv_ret == XMA_SUCCESS)
              break;
             else
               continue;
           } else {
             if (recv_ret == XMA_TRY_AGAIN) {
               continue;
             } else {
               break;
             }
           }
       }
     data_used = 0;

   }

    return data_used;
}

static const AVClass mpsoc_vcu_h264_class = {
    .class_name    =    "MPSOC H.264 decoder",
    .item_name     =    av_default_item_name,
    .option        =    options,
    .version       =    LIBAVUTIL_VERSION_INT,
};

AVCodec ff_h264_vcu_mpsoc_decoder = {
    .name                =    "mpsoc_vcu_h264",
    .long_name           =    NULL_IF_CONFIG_SMALL("MPSOC H.264 Decoder"),
    .type                =    AVMEDIA_TYPE_VIDEO,
    .id                  =    AV_CODEC_ID_H264,
    .init                =    mpsoc_vcu_decode_init,
    .decode              =    mpsoc_vcu_decode,
    .flush               =    mpsoc_vcu_flush,
    .bsfs                =    "h264_mp4toannexb",
    .close               =    mpsoc_vcu_decode_close,
    .priv_data_size      =    sizeof(mpsoc_vcu_dec_ctx),
    .priv_class          =    &mpsoc_vcu_h264_class,
    .capabilities        =    AV_CODEC_CAP_DELAY | AV_CODEC_CAP_AVOID_PROBING,
    .pix_fmts            =    (const enum AVPixelFormat[]) { AV_PIX_FMT_NV12,
                                                             AV_PIX_FMT_VCU_NV12,
                                                             AV_PIX_FMT_NONE
                                                           },
};

static const AVClass mpsoc_vcu_hevc_class = {
    .class_name    =    "MPSOC HEVC decoder",
    .item_name     =    av_default_item_name,
    .option        =    options,
    .version       =    LIBAVUTIL_VERSION_INT,
};

AVCodec ff_hevc_vcu_mpsoc_decoder = {
    .name                =    "mpsoc_vcu_hevc",
    .long_name           =    NULL_IF_CONFIG_SMALL("MPSOC HEVC Decoder"),
    .type                =    AVMEDIA_TYPE_VIDEO,
    .id                  =    AV_CODEC_ID_HEVC,
    .init                =    mpsoc_vcu_decode_init,
    .decode              =    mpsoc_vcu_decode,
    .flush               =    mpsoc_vcu_flush,
    .bsfs                =    "hevc_mp4toannexb",
    .close               =    mpsoc_vcu_decode_close,
    .priv_data_size      =    sizeof(mpsoc_vcu_dec_ctx),
    .priv_class          =    &mpsoc_vcu_hevc_class,
    .capabilities        =    AV_CODEC_CAP_DELAY | AV_CODEC_CAP_AVOID_PROBING,
    .pix_fmts            =    (const enum AVPixelFormat[]) { AV_PIX_FMT_NV12,
                                                             AV_PIX_FMT_VCU_NV12,
                                                             AV_PIX_FMT_NONE
                                                           },
};
